-- processor.vhd
-- Lab 5
-- Team CART
-- CSE 132L
-- The processor block is the top-level block in the processor design.
-- It houses the controller block, the instruction memory, the program
-- counter, and the datapath component, which contains all of the other
-- blocks in a single-cycle processor. It is responsible for connecting:
--   output of the program counter to input of the instruction memory
--   output of the instruction memory to input of the datapath
--   output of the datapath to input of the controller
--   output of the controller to input of the datapath
-- The processor block also receives the clock and reset signals for
-- all of the blocks in the design.

LIBRARY ieee;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_1164.ALL;
USE work.lab5_pack.ALL;

ENTITY processor IS
PORT (
	clk : IN std_logic;
	reset   : IN std_logic
);
END processor;

ARCHITECTURE Structural OF processor IS

	-- COMPONENT DECLARATION: HAZARD
	COMPONENT hazard IS
	PORT (
        MemtoRegM : IN std_logic;
		MemtoRegE : IN std_logic;
		
		RsD :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RsE :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RtD :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RtE :       IN std_logic_vector (NREG-1 DOWNTO 0);
        WriteRegE : IN std_logic_vector (NREG-1 DOWNTO 0);
        WriteRegM : IN std_logic_vector (NREG-1 DOWNTO 0);
		BranchD:    IN std_logic;
		JumpD:    IN std_logic;
	
		FlushE : OUT std_logic;
		StallD : OUT std_logic;
		StallF : OUT std_logic
	);
	END COMPONENT;

	-- COMPONENT DECLARATION: FORWARD
	COMPONENT forward IS
	PORT (
		RegWriteM : IN std_logic;
		RegWriteW : IN std_logic;
		RsD :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RtD :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RsE :       IN std_logic_vector (NREG-1 DOWNTO 0);
		RtE :       IN std_logic_vector (NREG-1 DOWNTO 0);
		WriteRegM : IN std_logic_vector (NREG-1 DOWNTO 0);
		WriteRegW : IN std_logic_vector (NREG-1 DOWNTO 0);
	
		ForwardAD : OUT std_logic_vector (1 DOWNTO 0);
		ForwardBD : OUT std_logic_vector (1 DOWNTO 0);
		ForwardAE : OUT std_logic_vector (1 DOWNTO 0);
		ForwardBE : OUT std_logic_vector (1 DOWNTO 0)
	);
	END COMPONENT;

	-- COMPONENT DECLARATION: FETCH
	COMPONENT stage_fetch is
	port (
	    clk : in std_logic;
        rst : in std_logic;
        StallF : in std_logic;
	    PC_p: in std_logic_vector(NBIT-1 downto 0);
	    InstrF : out std_logic_vector(NBIT-1 downto 0);
	    PCPlus4F : out std_logic_vector(NBIT-1 downto 0)

	);
	END COMPONENT;

	-- COMPONENT DECLARATION: DECODE
    component stage_decode is
    port (
    	clk : in std_logic;
    	rst : in std_logic;
    	StallD : in std_logic;
    	InstrF : in std_logic_vector(NBIT-1 downto 0);
    	PCPlus4F : in std_logic_vector(NBIT-1 downto 0);
    	RegWriteW : in std_logic;
    	WriteRegW : in std_logic_vector(NREG-1 downto 0);
    	ResultW : in std_logic_vector(NBIT-1 downto 0);
    	ForwardAD : in std_logic_vector(1 downto 0);
    	ForwardBD : in std_logic_vector(1 downto 0);
	    ForwardDataM : in std_logic_vector(NBIT-1 downto 0);
	    ForwardDataW : in std_logic_vector(NBIT-1 downto 0);
	    RegWriteD : out std_logic;
	    MemtoRegD : out std_logic;
    	MemWriteD : out std_logic;
	    BranchD : out std_logic;
	    ALUControlD : out std_logic_vector(NALU-1 downto 0);
    	ALUSrcD : out std_logic;
	    RegDstD : out std_logic;
    	SrcAD : out std_logic_vector(NBIT-1 downto 0);
	    SrcBD : out std_logic_vector(NBIT-1 downto 0);
    	RsD : out std_logic_vector(NREG-1 downto 0);
	    RtD : out std_logic_vector(NREG-1 downto 0);
    	RdD : out std_logic_vector(NREG-1 downto 0);
	    SignImmD : out std_logic_vector(NBIT-1 downto 0);
	    PCSrcD : out std_logic;
	    PCBranchD : out std_logic_vector(NBIT-1 downto 0);
	    PCJumpD : out std_logic;
	    PCJumpAddr : out std_logic_vector(NBIT-1 downto 0)
    );
    end component;

	-- COMPONENT DECLARATION: EXECUTE

	COMPONENT stage_execute IS
	PORT (
		clk : IN std_logic;
		rst : IN std_logic;

		ALUSrcIn :   IN std_logic;
		FlushE :     IN std_logic;
		MemtoRegIn : IN std_logic;
		MemWriteIn : IN std_logic;
		RegDstIn :   IN std_logic;
		RegWriteIn : IN std_logic;

		ALUControlIn : IN std_logic_vector (NALU-1 DOWNTO 0);
		Data1In :      IN std_logic_vector (NBIT-1 DOWNTO 0);
		Data2In :      IN std_logic_vector (NBIT-1 DOWNTO 0);
		ForwardAE :    IN std_logic_vector (1      DOWNTO 0);
		ForwardBE :    IN std_logic_vector (1      DOWNTO 0);
		ForwardDataM : IN std_logic_vector (NBIT-1 DOWNTO 0);
		ForwardDataW : IN std_logic_vector (NBIT-1 DOWNTO 0);
		RdIn :         IN std_logic_vector (NREG-1 DOWNTO 0);
		RsIn :         IN std_logic_vector (NREG-1 DOWNTO 0);
		RtIn :         IN std_logic_vector (NREG-1 DOWNTO 0);
		SignImmIn :    IN std_logic_vector (NBIT-1 DOWNTO 0);
		
		MemtoRegOut : OUT std_logic;
		MemWriteOut : OUT std_logic;
		RegWriteOut : OUT std_logic;
		
		ALUOut :       OUT std_logic_vector (NBIT-1 DOWNTO 0);
		RsOut :        OUT std_logic_vector (NREG-1 DOWNTO 0);
		RtOut :        OUT std_logic_vector (NREG-1 DOWNTO 0);
		WriteDataOut : OUT std_logic_vector (NBIT-1 DOWNTO 0);
		WriteRegOut :  OUT std_logic_vector (NREG-1 DOWNTO 0)
	);
	END COMPONENT;

	-- COMPONENT DECLARATION: MEMORY
	COMPONENT stage_memory IS
	PORT (
		clk : IN std_logic;
		rst : IN std_logic;
		RegWriteE : IN std_logic;
		MemtoRegE : IN std_logic;
		MemWriteE : IN std_logic;
		ALUOutE : IN std_logic_vector(NBIT-1 DOWNTO 0);
		WriteDataE : IN std_logic_vector(NBIT-1 DOWNTO 0);
		WriteRegE : IN std_logic_vector(NREG-1 DOWNTO 0);

		RegWriteM : OUT std_logic;
		MemtoRegM : OUT std_logic;
		ALUOutM : OUT std_logic_vector(NBIT-1 DOWNTO 0);
		WriteRegM : OUT std_logic_vector(NREG-1 DOWNTO 0);
		ReadDataM : OUT std_logic_vector(NBIT-1 DOWNTO 0);
		WriteRegM_hazard : OUT std_logic_vector(NREG-1 DOWNTO 0);
		ALUOutM_execute : OUT std_logic_vector(NBIT-1 DOWNTO 0)
	);
	END COMPONENT;

	-- COMPONENT DECLARATION: WRITEBACK

	COMPONENT stage_writeback is
	port (
	    clk : in std_logic;
        rst : in std_logic;
	    RegWriteM: in std_logic;
	    MemtoRegM : in std_logic;
	    ALUOutM : in std_logic_vector(NBIT-1 downto 0);
	    ReadDataM : in std_logic_vector(NBIT-1 downto 0);
	    WriteRegM : in std_logic_vector(NREG-1 downto 0);
        PCSrcD : in std_logic;
		PCJumpD : in std_logic;
		PCJumpAddr : in std_logic_vector(NBIT-1 downto 0);
	    PCBranchD : in std_logic_vector(NBIT-1 downto 0);
	    PCPlus4F : in std_logic_vector(NBIT-1 downto 0);
        RegWriteW : out std_logic;
	    WriteRegW: out std_logic_vector(NREG-1 downto 0);
	    ResultW : out std_logic_vector(NBIT-1 downto 0);
	    PC_p : out std_logic_vector(NBIT-1 downto 0)

	);
	END COMPONENT;

	-- SIGNAL DECLARATION: HAZARD

	SIGNAL FlushE : std_logic;
	SIGNAL StallD : std_logic;
	SIGNAL StallF : std_logic;

	-- SIGNAL DECLARATION: FORWARD
	
	SIGNAL ForwardAD : std_logic_vector (1 DOWNTO 0);
	SIGNAL ForwardBD : std_logic_vector (1 DOWNTO 0);
	SIGNAL ForwardAE : std_logic_vector (1 DOWNTO 0);
	SIGNAL ForwardBE : std_logic_vector (1 DOWNTO 0);

	-- SIGNAL DECLARATION: FETCH

	SIGNAL InstrF :   std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL PC :       std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL PCPlus4F : std_logic_vector (NBIT-1 DOWNTO 0);

	-- SIGNAL DECLARATION: DECODE

	SIGNAL ALUSrcD :   std_logic;
	SIGNAL BranchD :   std_logic;
	SIGNAL MemtoRegD : std_logic;
	SIGNAL MemWriteD : std_logic;
	SIGNAL RegDstD :   std_logic;
	SIGNAL RegWriteD : std_logic;
	SIGNAL PCJumpD :   std_logic;
	SIGNAL PCSrcD :    std_logic;

	SIGNAL ALUControlD : std_logic_vector (NALU-1 DOWNTO 0);
	SIGNAL InstrD :      std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL PCBranchD :   std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL PCJumpAddr :  std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL PCPlus4D :    std_logic_vector (NBIT-1 DOWNTO 0);
    SIGNAL RsD :         std_logic_vector (NREG-1 DOWNTO 0);
    SIGNAL RtD :         std_logic_vector (NREG-1 DOWNTO 0);
    SIGNAL RdD :         std_logic_vector (NREG-1 DOWNTO 0);
	SIGNAL SignImmD :    std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL SrcAD :       std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL SrcBD :       std_logic_vector (NBIT-1 DOWNTO 0);

	-- SIGNAL DECLARATION: EXECUTE

	SIGNAL ALUSrcE :   std_logic;
	SIGNAL MemtoRegE : std_logic;
	SIGNAL MemWriteE : std_logic;
	SIGNAL RegDstE :   std_logic;
	SIGNAL RegWriteE : std_logic;
 
	SIGNAL ALUControlE :  std_logic_vector (NALU-1 DOWNTO 0);
	SIGNAL ALUOutE :      std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL ForwardDataM : std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL ForwardDataW : std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL RsE :          std_logic_vector (NREG-1 DOWNTO 0);
	SIGNAL RtE :          std_logic_vector (NREG-1 DOWNTO 0);
	SIGNAL SignImmE :     std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL WriteDataE :   std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL WriteRegE :    std_logic_vector (NREG-1 DOWNTO 0);

	-- SIGNAL DECLARATION: MEMORY

	SIGNAL MemtoRegM : std_logic;
	SIGNAL MemWriteM : std_logic;
	SIGNAL RegWriteM : std_logic;

	SIGNAL ALUOutM :    std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL ReadDataM :  std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL WriteDataM : std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL WriteRegM :  std_logic_vector (NREG-1 DOWNTO 0);

	-- SIGNAL DECLARATION: WRITEBACK

	SIGNAL MemtoRegW : std_logic;
	SIGNAL RegWriteW : std_logic;

	SIGNAL ALUOutW :   std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL ReadDataW : std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL ResultW :   std_logic_vector (NBIT-1 DOWNTO 0);
	SIGNAL WriteRegW : std_logic_vector (NREG-1 DOWNTO 0);

BEGIN

	-- COMPONENT INITIALIZATION: HAZARD
	hazard_unit : hazard
	PORT MAP (
        MemtoRegM => MemtoRegM,
		MemtoRegE => MemtoRegE,
		RsD       => RsD,
		RsE       => RsE,
		RtD       => RtD,
		RtE       => RtE,
		WriteRegE => WriteRegE,
        WriteRegM => WriteRegM,
		BranchD   => BranchD,
		JumpD     => PCJumpD,
	
		FlushE => FlushE,
		StallD => StallD,
		StallF => StallF
	);

	-- COMPONENT INITIALIZATION: FORWARD
	forward_unit : forward
	PORT MAP (
		RegWriteM => RegWriteM,
		RegWriteW => RegWriteW,
		RsD       => RsD,
		RtD       => RtD,
		RsE       => RsE,
		RtE       => RtE,
		WriteRegM => WriteRegM,
		WriteRegW => WriteRegW,
		
		ForwardAD => ForwardAD,
		ForwardBD => ForwardBD,
		ForwardAE => ForwardAE,
		ForwardBE => ForwardBE
	);

	-- COMPONENT INITIALIZATION: FETCH
		fetch_stage : stage_fetch
	  PORT MAP(
	  clk => clk,
      rst => reset,
      StallF => StallF, 
	  PC_p => PC,
	  InstrF => InstrF,
	  PCPlus4F => PCPlus4F

	  );

	-- COMPONENT INITIALIZATION: DECODE
    decode_stage : stage_decode
    port map (
        clk => clk,
        rst => reset,
        StallD => StallD,
        InstrF => InstrF,
        PCPlus4F => PCPlus4F,
        RegWriteW => RegWriteW,
        WriteRegW => WriteRegW,
        ResultW => ResultW,
        ForwardAD => ForwardAD,
        ForwardBD => ForwardBD,
        ForwardDataM => ForwardDataM,
        ForwardDataW => ForwardDataW,
        RegWriteD => RegWriteD,
        MemtoRegD => MemtoRegD,
        MemWriteD => MemWriteD,
        BranchD => BranchD,
        ALUControlD => ALUControlD,
        ALUSrcD => ALUSrcD,
        RegDstD => RegDstD,
        SrcAD => SrcAD,
        SrcBD => SrcBD,
        RsD => RsD,
        RtD => RtD,
        RdD => RdD,
        SignImmD => SignImmD,
		PCSrcD => PCSrcD,
		PCBranchD => PCBranchD,
		PCJumpD => PCJumpD,
		PCJumpAddr => PCJumpAddr
    );

	-- COMPONENT INITIALIZATION: EXECUTE

	execute_stage : stage_execute
	PORT MAP (
		clk => clk,
		rst => reset,

		ALUSrcIn   => ALUSrcD,
		FlushE     => FlushE,
		MemtoRegIn => MemtoRegD,
		MemWriteIn => MemWriteD,
		RegDstIn   => RegDstD,
		RegWriteIn => RegWriteD,

		ALUControlIn => ALUControlD,
		Data1In      => SrcAD,
		Data2In      => SrcBD,
		ForwardAE    => ForwardAE,
		ForwardBE    => ForwardBE,
		ForwardDataM => ForwardDataM,
		ForwardDataW => ResultW,
		RdIn         => RdD,
		RsIn         => RsD,
		RtIn         => RtD,
		SignImmIn    => SignImmD,

		MemtoRegOut => MemtoRegE,
		MemWriteOut => MemWriteE,
		RegWriteOut => RegWriteE,

		ALUOut       => ALUOutE,
		RsOut        => RsE,
		RtOut        => RtE,
		WriteDataOut => WriteDataE,
		WriteRegOut  => WriteRegE
	);

	-- COMPONENT INITIALIZATION: MEMORY
	memory_stage : stage_memory
	PORT MAP (
		clk => clk,
		rst => reset,
		
		RegWriteE  => RegWriteE,
		MemtoRegE  => MemtoRegE,
		MemWriteE  => MemWriteE,
		ALUOutE    => ALUOutE,
		WriteDataE => WriteDataE,
		WriteRegE  => WriteRegE,

		RegWriteM        => RegWriteM,
		MemtoRegM        => MemtoRegM,
		ALUOutM          => ALUOutM,
		WriteRegM        => WriteRegM,
		ReadDataM        => ReadDataM,
		WriteRegM_hazard => WriteRegM,
		ALUOutM_execute  => ForwardDataM
	);

	-- COMPONENT INITIALIZATION: WRITEBACK
	writeback_stage : stage_writeback
	PORT MAP (
		clk => clk,
        rst => reset,
		RegWriteM => RegWriteM,
		MemtoRegM => MemtoRegM,
		ALUOutM => ALUOutM,
		ReadDataM => ReadDataM,
		WriteRegM => WriteRegM,
        PCSrcD => PCSrcD,
		PCJumpD => PCJumpD,
		PCJumpAddr => PCJumpAddr,
		PCBranchD => PCBranchD,
		PCPlus4F => PCPlus4F,
		RegWriteW => RegWriteW,
		WriteRegW => WriteRegW,
		ResultW => ResultW,
		PC_p => PC
	);

END Structural;
