-- datapath.vhd
-- Lab 2
-- Team CART
-- CSE 132L
-- The datapath component is responsible for connecting the following
-- blocks in the datapath: Register file, ALU, and data memory. It takes
-- the 32-bit instruction, and uses bit slicing to send the func and op
-- fields of the instruction to the controller, the read and write addresses
-- for the register file, and the 16-bit immediate value to the input
-- of the sign extender. It also receives the output of the controller
-- and connects them to the following inputs:
--   The function code of the ALU
--   The select lines of the multiplexers
--   The write enable lines for the register file and data memory
-- This block also connects the output of the multiplexers to the correct
-- inputs for the register file and ALU.

----------------------------------------------------------------
-- Instruction formats and references may be found on pages 119
-- and 120 of the 5th Edition textbook.
----------------------------------------------------------------

LIBRARY ieee;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_1164.ALL;
use work.lab2_pack.all;

ENTITY datapath IS
	PORT (
		-- inputs from processor
		clk :  IN std_logic;	-- clock
        rst :  IN std_logic;    -- reset
		inst : IN std_logic_vector(NBIT-1 DOWNTO 0);	-- instruction
		
		-- inputs from controller
		ci_alu :	IN std_logic_vector(NALU-1 DOWNTO 0);	-- ALU function code
		ci_br : 	IN std_logic;	-- if 1, next instruction is branch
		ci_mwe : 	IN std_logic;	-- write enable bit for memory
		ci_rwe :	IN std_logic;	-- write enable bit for register file
		ci_sel_am :	IN std_logic;	-- selector between ALU result and memory result
		ci_sel_dt :	IN std_logic;	-- selector between rd and rt as destination register
		ci_sel_ri :	IN std_logic;	-- selector between register and immediate value for ALU op
	
		-- outputs to controller
		co_o : OUT std_logic_vector(NOPC-1 DOWNTO 0);	-- opcode bits to controller
		co_f : OUT std_logic_vector(NALU-1 DOWNTO 0)	-- ALU function bits to controller
	);
END datapath;

ARCHITECTURE Behavioral OF datapath IS

	-- datapath components

	COMPONENT alu IS
		PORT (
			Func_in :		IN std_logic_vector (NALU-1 DOWNTO 0);
			A_in : 			IN std_logic_vector (NBIT-1 DOWNTO 0);
			B_in : 			IN std_logic_vector (NBIT-1 DOWNTO 0);
			O_out : 		OUT std_logic_vector (NBIT-1 DOWNTO 0);
			Branch_out : 	OUT std_logic;
			Jump_out : 		OUT std_logic
		);
	END COMPONENT;

	COMPONENT mux IS
        GENERIC (
            NBIT :	integer := 1	-- number of bits in each input
        );

		PORT (
			s : IN std_logic;
			a :	IN std_logic_vector(NBIT-1 DOWNTO 0);
			b :	IN std_logic_vector(NBIT-1 DOWNTO 0);
			o :	OUT std_logic_vector(NBIT-1 DOWNTO 0) 	
		);
	END COMPONENT;

	COMPONENT ram IS
		PORT (
			clk : 	IN std_logic;
			we : 	IN std_logic;
			addr : 	IN std_logic_vector (NMEM-1 DOWNTO 0);
			dataI : IN std_logic_vector (NBIT-1 DOWNTO 0);
			dataO : OUT std_logic_vector (NBIT-1 DOWNTO 0)
		);
	END COMPONENT;

	COMPONENT regfile IS
		PORT (
			clk :		IN std_logic;							-- clock
			rst_s : 	IN std_logic;							-- synchronous reset
			we : 		IN std_logic;							-- write enable
			raddr_1 : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- read address 1
			raddr_2 : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- read address 2
			waddr : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- write address
			rdata_1 : 	OUT std_logic_vector (NBIT-1 DOWNTO 0); -- read data 1
			rdata_2 : 	OUT std_logic_vector (NBIT-1 DOWNTO 0); -- read data 2
			wdata : 	IN std_logic_vector (NBIT-1 DOWNTO 0)	-- write data 1
    	);
	END COMPONENT;

	COMPONENT sign_ext IS
		PORT(
			input : IN std_logic_vector(NIMM-1 DOWNTO 0);
			output : OUT std_logic_vector(NBIT-1 DOWNTO 0)
		);
	END COMPONENT;

	-- instruction splitting
	SIGNAL inst_funct :	std_logic_vector(NALU-1 DOWNTO 0);
	SIGNAL inst_imm :	std_logic_vector(NIMM-1 DOWNTO 0);
	SIGNAL inst_opc : 	std_logic_vector(NOPC-1 DOWNTO 0);
	SIGNAL inst_rd : 	std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL inst_rs : 	std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL inst_rt : 	std_logic_vector(NREG-1 DOWNTO 0);

	-- register file IO
	SIGNAL reg_rd1 : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL reg_rd2 : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL reg_wa :  std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL reg_wd :  std_logic_vector(NBIT-1 DOWNTO 0);
	
	-- sign extension output
	SIGNAL imm_ext : std_logic_vector(NBIT-1 DOWNTO 0);

	-- ALU IO
	SIGNAL alu_opB :    std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL alu_res : std_logic_vector(NBIT-1 DOWNTO 0);
	
	-- memory IO
	SIGNAL mem_d : std_logic_vector(NBIT-1 DOWNTO 0);

BEGIN
	datapath_reg: regfile
		PORT MAP	(clk     => clk,
					 rst_s   => rst,
					 we      => ci_rwe,
					 raddr_1 => inst_rs,
					 raddr_2 => inst_rt,
					 waddr   => reg_wa,
					 rdata_1 => reg_rd1,
					 rdata_2 => reg_rd2,
					 wdata   => reg_wd);

	datapath_mux_alub : mux
		GENERIC MAP (NBIT => 32)
		PORT MAP	(s    => ci_sel_ri,
					 a    => reg_rd2,
					 b    => imm_ext,
					 o    => alu_opB);

	datapath_alu : alu
		PORT MAP (Func_in    => ci_alu,
				  A_in       => reg_rd1,
				  B_in       => alu_opB,
				  O_out      => alu_res);
				  -- Branch_out => 
				  -- Jump_out   => );

	datapath_ram : ram
		PORT MAP	(clk   => clk,
					 we    => ci_mwe,
					 addr  => alu_res(8 DOWNTO 0),
					 dataI => reg_rd2,
					 dataO => mem_d);

	datapath_mux_result : mux
    GENERIC MAP (NBIT => NBIT)
		PORT MAP	(s    => ci_sel_am,
					 a    => alu_res,
					 b    => mem_d,
					 o    => reg_wd);

	datapath_reg_sel : mux
    GENERIC MAP (NBIT => NREG)
		PORT MAP	(s    => ci_sel_dt,
					 a    => inst_rt,
					 b    => inst_rd,
					 o	  => reg_wa);
                     
    sign_extender : sign_ext
        PORT MAP    (input => inst_imm,
                     output => imm_ext);

	-- splitting the instruction
	inst_funct	<= inst(NALU-1 DOWNTO 0);
	inst_imm 	<= inst(NIMM-1 DOWNTO 0);
	inst_opc	<= inst(31 DOWNTO 26);
	inst_rd 	<= inst(15 DOWNTO 11);
	inst_rs 	<= inst(25 DOWNTO 21);
	inst_rt 	<= inst(20 DOWNTO 16);

	-- outputs to controller
	co_o <= inst_opc;
	co_f <= inst_funct;
 
END Behavioral; 
