-- controller.vhd
-- Lab 3
-- Team CART
-- CSE 132L
-- The controller block is responsible for setting the correct value
-- for the select lines of the following multiplexers in the datapath:
--   Choosing the write address for the register file
--   Choosing the write data for the register file
--   Choosing the second operand of the ALU
-- The controller also determines whether the write enable for the
-- register file and the data memory should be enabled or not, based
-- on the instruction, as determined by the func and op fields of the
-- instruction. Additionally, it determines the correct function code
-- for the ALU from func and op.
-- 
-- Notes:
--     regwrite and memwrite are 1 only for instructions that need to
--      register file or memory, and only those that are supported.
--     alusrca selects between read data 1 (0) and shift amount, for lui,
--      sll, srl, and sra (1).
--     alusrcb selects between read data 2 (0) and sign-extended immediate
--      value (1).
--     memtoreg selects between ALU output (0) and memory output (1).
--     regwrite enables write for register file (R-type, I-type, load)
--     memwrite enables write for memory (store)
--     shamt specifies amount of shift for sll, srl, and sra
--      (inst(10 downto 6)) and lui (0x10).

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use work.lab3_pack.all;

entity controller is
    port (
	inst : in std_logic_vector(NBIT-1 downto 0); -- Instruction
	raddr1 : out std_logic_vector(NREG-1 downto 0);
	raddr2 : out std_logic_vector(NREG-1 downto 0);
	waddr : out std_logic_vector(NREG-1 downto 0);
	imm : out std_logic_vector(NIMM-1 downto 0);
	shamt : out std_logic_vector(NBIT-1 downto 0);
	jump_imm : out std_logic_vector(25 downto 0);
	alusrca : out std_logic;
	alusrcb : out std_logic;
	wdatasrc : out std_logic;
        alucontrol : out std_logic_vector(NALU-1 downto 0); -- ALU funct code
        memtoreg : out std_logic;
        regwrite : out std_logic;
        memwrite : out std_logic;
        branch : out std_logic;
	jump : out std_logic;
	dsize : out std_logic_vector(1 downto 0);
	dsgnd : out std_logic;
	jtype : out std_logic
    );
end controller;

architecture Behavioral of controller is

signal op_imm : std_logic_vector(NALU-1 downto 0);
signal func_mod : std_logic_vector(NALU-1 downto 0);

begin
    raddr1 <= inst(25 downto 21); -- rs

    raddr2 <= inst(20 downto 16); -- rt

    waddr <= inst(20 downto 16) when inst(31 downto 29) = "001" -- I-type
				  or inst(31 downto 29) = "100" -- load
	else "11111" when inst(31 downto 26) = "000011" -- jal
	else inst(15 downto 11);

    imm <= inst(NIMM-1 downto 0); -- 16-bit immediate

    -- Applies only to lui, sll, srl, and sra
    shamt <= (4 => '1', others => '0') when inst(31 downto 26) = "001111" -- lui
            else "000000000000000000000000000" & Inst(10 downto 6); -- shift

    jump_imm <= inst(25 downto 0); -- 26-bit immediate

    -- alusrca is 1 for for lui, sll, srl or sra; 0 for other instructions
    alusrca <= '1' when inst(31 downto 26) = "001111"
		     or (inst(31 downto 26) = "000000"
                        and inst(5 downto 2) = "0000")
          else '0';

    -- alusrcb is 0 for R-type, branch or jump; 1 for I-type, load, or store
    alusrcb <= '0' when inst(31 downto 29) = "000" else '1';

    wdatasrc <= '1' when inst(31 downto 26) = "000011" -- jal
		      or (inst(31 downto 26) = "000000"
			  and inst(5 downto 0) = "001001") -- jalr
	   else '0';

    with inst(28 downto 26) select
        op_imm <= "100000" when "000", -- addi (add)
                  "100001" when "001", -- addiu (addu)
                  "101000" when "010", -- slti (slt)
                  "101001" when "011", -- sltiu (sltu)
                  "100100" when "100", -- andi (and)
                  "100101" when "101", -- ori (or)
                  "100110" when "110", -- xori (xor)
                  "000000" when "111", -- lui (sll)
                  (others => '0') when others;

    -- Reconcile differences in opcodes for slt/sltu between book and lab
    with inst(5 downto 0) select
	func_mod <= "101000" when "101010",
		    "101001" when "101011",
		    inst(5 downto 0) when others;

    alucontrol <= "111000" when inst(31 downto 26) = "000001"
                                and inst(20 downto 16) = "00000" -- bltz
             else "111001" when inst(31 downto 26) = "000001"
                                and inst(20 downto 16) = "00001" -- bgez
             else "111100" when inst(31 downto 26) = "000100" -- beq
             else "111101" when inst(31 downto 26) = "000101" -- bne
             else "111110" when inst(31 downto 26) = "000110" -- blez
             else "111111" when inst(31 downto 26) = "000111" -- bgtz
             else op_imm when inst(31 downto 29) = "001" -- I-type
             else "100000" when inst(31 downto 30) = "10" -- load/store
             else func_mod; -- others

    -- memtoreg is 1 for load; 0 for other instructions
    memtoreg <= '1' when inst(31 downto 29) = "100" else '0';

    -- regwrite is 1 for R-type, I-type, load, jal or jalr; 0 for others
    regwrite <= '1' when (inst(31 downto 26) = "000000"
                          and ((inst(5 downto 3) = "000" -- Shifts
                                and inst(1 downto 0) /= "01") -- Exclude invalid
                               or inst(5 downto 0) = "001001" -- jalr
                               or inst(5 downto 3) = "100" -- add/sub/logic
                               or inst(5 downto 1) = "10101")) -- slt/sltu
                         or inst(31 downto 26) = "000011" -- jal
                         or inst(31 downto 29) = "001" -- Immediate
                         or (inst(31 downto 29) = "100" -- Load
                             and inst(27 downto 26) /= "10" -- Exclude lwl/lwr
                             and inst(28 downto 26) /= "111") -- Exclude invalid
           else '0';

    -- memwrite is 1 for store; 0 for other instructions
    memwrite <= '1' when inst(31 downto 27) = "10100" -- sb/sh
                         or inst(31 downto 26) = "101011" -- sw
           else '0';

    branch <= '1' when inst(31 downto 28) = "0001" -- beq/bne/blez/bgtz
                       or (inst(31 downto 26) = "000001"
                           and inst(20 downto 17) = "0000") -- bltz/bgez
         else '0';

    jump <= '1' when inst(31 downto 27) = "00001" -- jump/jal
                     or (inst(31 downto 26) = "000000"
                         and inst(5 downto 1) = "00100") -- jr/jalr
       else '0';

    dsize <= "00" when inst(31 downto 30) = "10"
		   and inst(27 downto 26) = "00" -- lb/lbu/sb
	else "01" when inst(31 downto 30) = "10"
		   and inst(27 downto 26) = "01" -- lh/lhu/sh
	else "11"; -- lw/sw

    dsgnd <= '0' when inst(31 downto 27) = "10010" else '1'; -- lbu/lhu

    jtype <= '1' when inst(31 downto 26) = "000000"
		  and inst(5 downto 1) = "00100" -- jr/jalr
	else '0';

end Behavioral;
