-- processor.vhd
-- Lab 2
-- Team CART
-- CSE 132L
-- The processor block is the top-level block in the processor design.
-- It houses the controller block, the instruction memory, the program
-- counter, and the datapath component, which contains all of the other
-- blocks in a single-cycle processor. It is responsible for connecting:
--   output of the program counter to input of the instruction memory
--   output of the instruction memory to input of the datapath
--   output of the datapath to input of the controller
--   output of the controller to input of the datapath
-- The processor block also receives the clock and reset signals for
-- all of the blocks in the design.
LIBRARY ieee;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_1164.ALL;
use work.lab2_pack.all;

ENTITY processor IS

	PORT (
		ref_clk	: IN std_logic;
		reset	: IN std_logic
	);
END processor;

ARCHITECTURE Structural OF processor IS

	-- component declaration

	COMPONENT controller IS
	    PORT (
			-- inputs from datapath
	        Func :	IN std_logic_vector(NALU-1 DOWNTO 0);
	        Op : 	IN std_logic_vector(NOPC-1 DOWNTO 0);
			
			-- outputs to datapath
			ALUControl : OUT std_logic_vector(NALU-1 DOWNTO 0);
	        ALUSrc : 	 OUT std_logic;
	        MemtoReg : 	 OUT std_logic;
	        RegWrite : 	 OUT std_logic;
	        RegDst : 	 OUT std_logic;
	        MemWrite : 	 OUT std_logic;
	        Branch : 	 OUT std_logic
	    );
	END COMPONENT;

	COMPONENT datapath IS

		PORT (
			-- inputs from processor
			clk :  IN std_logic;	-- clock
            rst :  IN std_logic;    -- reset
			inst : IN std_logic_vector(NBIT-1 DOWNTO 0);	-- instruction
			
			-- inputs from controller
			ci_alu :	IN std_logic_vector(NALU-1 DOWNTO 0);	-- ALU function code
			ci_br : 	IN std_logic;	-- if 1, next instruction is branch
			ci_mwe : 	IN std_logic;	-- write enable bit for memory
			ci_rwe : 	IN std_logic;	-- write enable bit for register file
			ci_sel_am :	IN std_logic;	-- selector between ALU result and memory result
			ci_sel_dt :	IN std_logic;	-- selector between rd and rt as destination register
			ci_sel_ri :	IN std_logic;	-- selector between register and immediate value for ALU op
	
			-- outputs to controller
			co_o : OUT std_logic_vector(NOPC-1 DOWNTO 0);	-- opcode bits to controller
			co_f : OUT std_logic_vector(NALU-1 DOWNTO 0)	-- ALU function bits to controller
		);
	END COMPONENT;


	COMPONENT prog_counter IS
		PORT (
 			clk : IN  STD_LOGIC;
			rst : IN  STD_LOGIC;
			output : OUT  STD_LOGIC_VECTOR (NBIT-1 downto 0)
        );
	END COMPONENT;
    
    COMPONENT rom IS
        PORT (
            addr   : IN std_logic_vector(NMEM-1 DOWNTO 0);
            dataIO : INOUT std_logic_vector(NBIT-1 DOWNTO 0)
        );
	END COMPONENT;

	-- signal declaration

    SIGNAL pc : std_logic_vector(NBIT-1 DOWNTO 0) := (others => '0');
	SIGNAL inst : std_logic_vector(NBIT-1 DOWNTO 0);
	
	SIGNAL c_alu :	  std_logic_vector(NALU-1 DOWNTO 0);
	SIGNAL c_br :	  std_logic;
	SIGNAL c_funct :  std_logic_vector(NALU-1 DOWNTO 0);
	SIGNAL c_mwe : 	  std_logic;
	SIGNAL c_rwe : 	  std_logic;
	SIGNAL c_opc :	  std_logic_vector(NOPC-1 DOWNTO 0);
	SIGNAL c_sel_am : std_logic;
	SIGNAL c_sel_dt : std_logic;
	SIGNAL c_sel_ri : std_logic;
	
BEGIN

	-- initialize components
	c : controller PORT MAP (
		Func => c_funct, 
		Op => c_opc,
		ALUControl => c_alu,
		ALUSrc => c_sel_ri,
		MemtoReg => c_sel_am,
		RegWrite => c_rwe,
		RegDst => c_sel_dt,
		MemWrite => c_mwe,
		Branch => c_br
	);

	d : datapath PORT MAP (
		clk => ref_clk,
        rst => reset,
		inst => inst,
		ci_alu => c_alu,
		ci_br => c_br,
		ci_mwe => c_mwe,
		ci_rwe => c_rwe,
		ci_sel_am => c_sel_am,
		ci_sel_dt => c_sel_dt,
		ci_sel_ri => c_sel_ri,
		co_o => c_opc,
		co_f => c_funct
	);
    
    prog_count : prog_counter PORT MAP (
        clk => ref_clk,
        rst => reset,
        output => pc
    );
    
    instr_mem : rom PORT MAP (
        addr => pc(NMEM-1 DOWNTO 0),
        dataIO => inst
    );

END Structural;
