library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use work.lab5_pack.all;

entity stage_decode is
port (
    clk : in std_logic;
    rst : in std_logic;
    StallD : in std_logic;
    InstrF : in std_logic_vector(NBIT-1 downto 0);
    PCPlus4F : in std_logic_vector(NBIT-1 downto 0);
    RegWriteW : in std_logic;
    WriteRegW : in std_logic_vector(NREG-1 downto 0);
    ResultW : in std_logic_vector(NBIT-1 downto 0);
    ForwardAD : in std_logic_vector(1 downto 0);
    ForwardBD : in std_logic_vector(1 downto 0);
    ForwardDataM : in std_logic_vector(NBIT-1 downto 0);
    ForwardDataW : in std_logic_vector(NBIT-1 downto 0);
    RegWriteD : out std_logic;
    MemtoRegD : out std_logic;
    MemWriteD : out std_logic;
    BranchD : out std_logic;
    ALUControlD : out std_logic_vector(NALU-1 downto 0);
    ALUSrcD : out std_logic;
    RegDstD : out std_logic;
    SrcAD : out std_logic_vector(NBIT-1 downto 0);
    SrcBD : out std_logic_vector(NBIT-1 downto 0);
    RsD : out std_logic_vector(NREG-1 downto 0);
    RtD : out std_logic_vector(NREG-1 downto 0);
    RdD : out std_logic_vector(NREG-1 downto 0);
    SignImmD : out std_logic_vector(NBIT-1 downto 0);
    PCSrcD : out std_logic;
    PCBranchD : out std_logic_vector(NBIT-1 downto 0);
    PCJumpD : out std_logic;
    PCJumpAddr : out std_logic_vector(NBIT-1 downto 0)
);
end stage_decode;

architecture Structural of stage_decode is

    signal InstrD : std_logic_vector(NBIT-1 downto 0);
    signal not_StallD : std_logic;
    signal SignImmTemp1 : std_logic_vector(NBIT-1 downto 0);
    signal SignImmTemp2 : std_logic_vector(NBIT-1 downto 0);
    signal PCPlus4D : std_logic_vector(NBIT-1 downto 0);
    signal ALUControl : std_logic_vector(NALU-1 downto 0);
    signal rdata1 : std_logic_vector(NBIT-1 downto 0);
    signal rdata2 : std_logic_vector(NBIT-1 downto 0);
    signal ZeroD : std_logic;
    signal t_BranchD : std_logic;
    signal comp1 : std_logic_vector(NBIT-1 downto 0);
    signal comp2 : std_logic_vector(NBIT-1 downto 0);
    signal jump_temp : std_logic_vector(27 downto 0);

    component fetch_decode_reg is
    port (
        clk : in std_logic;
        rst : in std_logic;
        en : in std_logic;
        InstrF : in std_logic_vector(NBIT-1 downto 0);
        PCPlus4F : in std_logic_vector(NBIT-1 downto 0);
        InstrD : out std_logic_vector(NBIT-1 downto 0);
        PCPlus4D : out std_logic_vector(NBIT-1 downto 0)
    );
    end component;
    
    component controller is
    port (
        Op : in std_logic_vector(NOPC-1 downto 0);
        Funct : in std_logic_vector(NALU-1 downto 0);
        RegWrite : out std_logic;
        MemtoReg : out std_logic;
        MemWrite : out std_logic;
        Branch : out std_logic;
        ALUControl : out std_logic_vector(NALU-1 downto 0);
        ALUSrc : out std_logic;
        RegDst : out std_logic;
        Jump : out std_logic
    );
    end component;

    component regfile is
    port (
        clk : in std_logic;
        rst_s : in std_logic; -- synchronous reset
        we : in std_logic; -- write enable
        raddr_1 : in std_logic_vector (NREG-1 downto 0); -- read address 1
        raddr_2 : in std_logic_vector (NREG-1 downto 0); -- read address 2
        waddr : in std_logic_vector (NREG-1 downto 0); -- write address
        rdata_1 : out std_logic_vector (NBIT-1 downto 0); -- read data 1
        rdata_2 : out std_logic_vector (NBIT-1 downto 0); -- read data 2
        wdata : in std_logic_vector (NBIT-1 downto 0) -- write data
    );
    end component;

    component sign_ext is
    port (
        imm : in std_logic_vector(NIMM-1 downto 0);
        imm_ext : out std_logic_vector(NBIT-1 downto 0)
    );
    end component;

    component adder is
    port (
        a : in std_logic_vector(NBIT-1 downto 0);
        b : in std_logic_vector(NBIT-1 downto 0);
        c : out std_logic_vector(NBIT-1 downto 0)
    );
    end component;

    component shifter is
    generic (
        SIZE : integer := 32;
        AMT : integer := 2
    );
    port (
        a : in std_logic_vector(SIZE-1 downto 0);
        o : out std_logic_vector(SIZE-1 downto 0)
    );
    end component;

    component comparator is
    port (
        Func_in : in std_logic_vector(NALU-1 downto 0);
        A_in : in std_logic_vector(NBIT-1 downto 0);
        B_in : in std_logic_vector(NBIT-1 downto 0);
        equal : out std_logic
    );
    end component;

    component mux3 is
    port (
        s : in  std_logic_vector (1      downto 0);
        a : in  std_logic_vector (NBIT-1 downto 0);
        b : in  std_logic_vector (NBIT-1 downto 0);
        c : in  std_logic_vector (NBIT-1 downto 0);
        o : out std_logic_vector (NBIT-1 downto 0)  
    );
    end component;

begin

    SrcAD <= rdata1;
    SrcBD <= rdata2;
    RsD <= InstrD(25 downto 21);
    RtD <= InstrD(20 downto 16);
    RdD <= InstrD(15 downto 11);

    not_StallD <= not StallD;
    SignImmD <= SignImmTemp1;
    ALUControlD <= ALUControl;
    BranchD <= t_BranchD;
    PCSrcD <= t_BranchD and ZeroD;
    jump_temp <= InstrD(25 downto 0) & "00";
    PCJumpAddr <= PCPlus4D(31 downto 28) & jump_temp;

    fetch_decode_pipeline : fetch_decode_reg
    port map (
        clk => clk,
        rst => rst,
        en => not_StallD,
        InstrF => InstrF,
        PCPlus4F => PCPlus4F,
        InstrD => InstrD,
        PCPlus4D => PCPlus4D
    );

    control : controller
    port map (
        Op => InstrD(31 downto 26),
        Funct => InstrD(5 downto 0),
        RegWrite => RegWriteD,
        MemtoReg => MemtoRegD,
        MemWrite => MemWriteD,
        Branch => t_BranchD,
        ALUControl => ALUControl,
        ALUSrc => ALUSrcD,
        RegDst => RegDstD,
        Jump => PCJumpD
    );

    register_file : regfile
    port map (
        clk => clk,
        rst_s => rst,
        we => RegWriteW,
        raddr_1 => InstrD(25 downto 21),
        raddr_2 => InstrD(20 downto 16),
        waddr => WriteRegW,
        rdata_1 => rdata1,
        rdata_2 => rdata2,
        wdata => ResultW
    );

    extend_imm : sign_ext
    port map (
        imm => InstrD(15 downto 0),
        imm_ext => SignImmTemp1
    );

    branch_shift : shifter
    generic map (
        SIZE => 32,
        AMT => 2
    )
    port map (
        a => SignImmTemp1,
        o => SignImmTemp2
    );

    branch_offset : adder
    port map (
        a => SignImmTemp2,
        b => PCPlus4D,
        c => PCBranchD
    );

    branch_compare : comparator
    port map (
        Func_in => ALUControl,
        A_in => comp1,
        B_in => comp2,
        equal => ZeroD
    );

	rdata1_mux3 : mux3
	PORT MAP (
		s => ForwardAD,
		a => rdata1,
		b => ForwardDataW,
		c => ForwardDataM,
		o => comp1
	);

	rdata2_mux3 : mux3
	PORT MAP (
		s => ForwardBD,
		a => rdata2,
		b => ForwardDataW,
		c => ForwardDataM,
		o => comp2
	);

end Structural;
