-- processor.vhd
-- Lab 3
-- Team CART
-- CSE 132L
-- The processor block is the top-level block in the processor design.
-- It houses the controller block, the instruction memory, the program
-- counter, and the datapath component, which contains all of the other
-- blocks in a single-cycle processor. It is responsible for connecting:
--   output of the program counter to input of the instruction memory
--   output of the instruction memory to input of the datapath
--   output of the datapath to input of the controller
--   output of the controller to input of the datapath
-- The processor block also receives the clock and reset signals for
-- all of the blocks in the design.

LIBRARY ieee;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_1164.ALL;
USE work.lab3_pack.ALL;

ENTITY processor IS
PORT (
	clk	: IN std_logic;
	reset	: IN std_logic
);
END processor;

ARCHITECTURE Structural OF processor IS

	-- component declaration
	COMPONENT dmem PORT (
		clk : in std_logic;
		we : in std_logic;
		a : in std_logic_vector(31 downto 0);
		wd : in std_logic_vector(31 downto 0);
		rd : out std_logic_vector(31 downto 0)
	);
	END COMPONENT;

	COMPONENT reg IS
	PORT (
		clk : IN  STD_LOGIC;
		rst : IN  STD_LOGIC;
		d : IN  STD_LOGIC_VECTOR (NBIT-1 downto 0);
		q : OUT STD_LOGIC_VECTOR (NBIT-1 downto 0)
        );
	END COMPONENT;
    
    	COMPONENT rom IS
        PORT (
        	addr   : IN std_logic_vector(NBIT-1 DOWNTO 0);
        	dataIO : INOUT std_logic_vector(NBIT-1 DOWNTO 0)
        );
	END COMPONENT;

	COMPONENT sign_ext IS
	PORT (
		input : IN std_logic_vector(NIMM-1 DOWNTO 0);
		output : OUT std_logic_vector(NBIT-1 DOWNTO 0)
	);
	END COMPONENT;

	COMPONENT controller IS
	PORT (
		-- inputs from datapath
	        inst :		IN std_logic_vector(NBIT-1 DOWNTO 0);	
		-- outputs to datapath
		raddr1 :	OUT std_logic_vector(NREG-1 DOWNTO 0);
		raddr2 :	OUT std_logic_vector(NREG-1 DOWNTO 0);
		waddr :		OUT std_logic_vector(NREG-1 DOWNTO 0);
		imm :		OUT std_logic_vector(NIMM-1 DOWNTO 0);
		shamt :		OUT std_logic_vector(NBIT-1 DOWNTO 0);
		jump_imm :	OUT std_logic_vector(25 DOWNTO 0);
		alusrca :	OUT std_logic;
		alusrcb :	OUT std_logic;
		wdatasrc :	OUT std_logic;
		alucontrol :	OUT std_logic_vector(NALU-1 DOWNTO 0);
		memtoreg :	OUT std_logic;
		regwrite :	OUT std_logic;
		memwrite :	OUT std_logic;
		branch :	OUT std_logic;
		jump :		OUT std_logic;
		dsize :		OUT std_logic_vector(1 downto 0);
		dsgnd :		OUT std_logic;
		jtype :		OUT std_logic
	);
	END COMPONENT;

	COMPONENT mux IS
	GENERIC (
        	NBIT :	integer := 1	-- number of bits in each input
	);
	PORT (
		s :	IN std_logic;
		a :	IN std_logic_vector(NBIT-1 DOWNTO 0);
		b :	IN std_logic_vector(NBIT-1 DOWNTO 0);
		o :	OUT std_logic_vector(NBIT-1 DOWNTO 0) 	
	);
	END COMPONENT;
	
	COMPONENT shifter IS
	GENERIC(
		SIZE : integer := 32;
		AMT : integer := 2
	);
	PORT (
		a : IN std_logic_vector(SIZE-1 DOWNTO 0);
		o : OUT std_logic_vector(SIZE-1 DOWNTO 0)
	);
	END COMPONENT;
		
	COMPONENT regfile IS
	PORT (
		clk :		IN std_logic;		-- clock
		rst_s : 	IN std_logic;		-- synchronous reset
		we : 		IN std_logic;		-- write enable
		raddr_1 : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- read address 1
		raddr_2 : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- read address 2
		waddr : 	IN std_logic_vector (NREG-1 DOWNTO 0);	-- write address
		rdata_1 : 	OUT std_logic_vector (NBIT-1 DOWNTO 0); -- read data 1
		rdata_2 : 	OUT std_logic_vector (NBIT-1 DOWNTO 0); -- read data 2
		wdata : 	IN std_logic_vector (NBIT-1 DOWNTO 0)	-- write data 1
    	);
	END COMPONENT;

	COMPONENT alu IS
	PORT (
		Func_in :	IN std_logic_vector (NALU-1 DOWNTO 0);
		A_in : 		IN std_logic_vector (NBIT-1 DOWNTO 0);
		B_in : 		IN std_logic_vector (NBIT-1 DOWNTO 0);
		O_out : 	OUT std_logic_vector (NBIT-1 DOWNTO 0);
		Branch_out : 	OUT std_logic
	);
	END COMPONENT;

	COMPONENT adder IS
	PORT(
		a:	IN STD_LOGIC_VECTOR(NBIT-1 DOWNTO 0);
		b:	IN STD_LOGIC_VECTOR(NBIT-1 DOWNTO 0);
		c:	OUT STD_LOGIC_VECTOR(NBIT-1 DOWNTO 0)
	);
	END COMPONENT;

	-- signal declaration

	SIGNAL pc :   std_logic_vector(NBIT-1 DOWNTO 0) := (others => '0');
	SIGNAL inst : std_logic_vector(NBIT-1 DOWNTO 0);
	
	SIGNAL c_raddr1 :  std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL c_raddr2 :  std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL c_waddr :  std_logic_vector(NREG-1 DOWNTO 0);
	SIGNAL c_imm :	  std_logic_vector(NIMM-1 DOWNTO 0);
	SIGNAL c_shamt :  std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL c_jimm  :  std_logic_vector(25 DOWNTO 0);
	SIGNAL c_alu_a  : std_logic;
	SIGNAL c_alu_b :  std_logic;
	SIGNAL c_wd_src : std_logic;
	SIGNAL c_alu :	  std_logic_vector(NALU-1 DOWNTO 0);
	SIGNAL c_br :	  std_logic;
	SIGNAL c_jump :   std_logic;
	SIGNAL c_mwe : 	  std_logic;
	SIGNAL c_rwe : 	  std_logic;
	SIGNAL c_dsize :  std_logic_vector(1 downto 0);
	SIGNAL c_dsgnd :  std_logic;
	SIGNAL c_j_imm_reg : std_logic;

	SIGNAL c_sel_am : std_logic;



	-- register file IO
	SIGNAL reg_rd1 : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL reg_rd2 : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL wd_result :  std_logic_vector(NBIT-1 DOWNTO 0);
	
	-- sign extension output
	SIGNAL imm_ext : std_logic_vector(NBIT-1 DOWNTO 0);

	-- ALU IO
	SIGNAL alu_opA : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL alu_opB : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL alu_res : std_logic_vector(NBIT-1 DOWNTO 0);
	SIGNAL alu_branch: std_logic;
	
	-- memory IO
	SIGNAL mem_d : std_logic_vector(NBIT-1 DOWNTO 0);
	
	-- shift left 2 (32) output
	SIGNAL sl2_32_output : std_logic_vector(NBIT-1 DOWNTO 0);
	
	--shift left 2 (28) output
	SIGNAL sl2_28_output : std_logic_vector(27 DOWNTO 0);

	--Adder pc+4 result
	SIGNAL adder_plus4_result : std_logic_vector(NBIT-1 DOWNTO 0);
        SIGNAL reg_wd : std_logic_vector(NBIT-1 DOWNTO 0);
	
	--Adder sl2_plus4 result 
	SIGNAL adder_result : std_logic_vector(NBIT-1 DOWNTO 0);

	-- branch Result
	SIGNAL branch_result : std_logic;	

	--branch mux result
	SIGNAL branch_mux_output : std_logic_vector(NBIT-1 DOWNTO 0);
	
	--jump mux result
	SIGNAL jump_mux_output : std_logic_vector(NBIT-1 DOWNTO 0);
	
	SIGNAL jump_imm_addr : std_logic_vector(NBIT-1 DOWNTO 0);

	SIGNAL jump_address : std_logic_vector(NBIT-1 DOWNTO 0);

	--Shifter Input + 2
	SIGNAL inst_input_plus2 : std_logic_vector(27 DOWNTO 0);
BEGIN

	--pre shifting inst[25:0] by 2 to increase size to 28
	inst_input_plus2 <= "00" & c_jimm;

	--branch AND
	branch_result <= c_br AND alu_branch;

	--jump address concatenation
	jump_imm_addr <= adder_plus4_result(31 DOWNTO 28) & sl2_28_output;
	
	-- initialize components
	prog_counter : reg
	PORT MAP (
        	clk => clk,
	        rst => reset,
		d => jump_mux_output,
        	q => pc
	);

	increment_pc : adder
	PORT MAP(
		a => pc,
		b => (2 => '1', others => '0'),
		c => adder_plus4_result
	);    

	branch_target_adder : adder
	PORT MAP(
		a => adder_plus4_result,
		b => sl2_32_output,
		c => adder_result
	);

	instr_mem : rom
	PORT MAP (
	        addr => pc,
	        dataIO => inst
	);
                     
    	sign_extender : sign_ext
	PORT MAP (
		input => c_imm,
                output => imm_ext
	);
	
	branch_sl2 : shifter
	GENERIC MAP( SIZE => 32,
		     AMT => 2 )
	PORT MAP (
		a => imm_ext,
		o => sl2_32_output
	);
	
	jump_sl2 : shifter
	GENERIC MAP( SIZE => 28,
		     AMT => 2 )
	PORT MAP(
		a => inst_input_plus2,
		o => sl2_28_output
	);

	c : controller
	PORT MAP (
		inst => inst,
		raddr1 => c_raddr1,
		raddr2 => c_raddr2,
		waddr => c_waddr,
		imm => c_imm,
		shamt => c_shamt,
		jump_imm => c_jimm,
		alusrca => c_alu_a,
		alusrcb => c_alu_b,
		wdatasrc => c_wd_src,
		alucontrol => c_alu,
		memtoreg => c_sel_am,
		regwrite => c_rwe,
		memwrite => c_mwe,
		branch => c_br,
		jump => c_jump,
		dsize => c_dsize,
		dsgnd => c_dsgnd,
		jtype => c_j_imm_reg
	);
	
	wdata_src_mux :mux
	GENERIC MAP(NBIT => 32)
	PORT MAP(
		s => c_wd_src,
		a => reg_wd,
		b => adder_plus4_result,
		o => wd_result
	);

	alu_a : mux
	GENERIC MAP(NBIT =>32)
	PORT MAP(
		s => c_alu_a,
		a => reg_rd1,
		b => c_shamt,
		o => alu_opA
	);

	alub_reg_imm : mux
	GENERIC MAP (NBIT => 32)
	PORT MAP (
		s    => c_alu_b,
		a    => reg_rd2,
		b    => imm_ext,
		o    => alu_opB
	);

	write_alu_mem : mux
	GENERIC MAP (NBIT => NBIT)
	PORT MAP (
		s    => c_sel_am,
		a    => alu_res,
		b    => mem_d,
		o    => reg_wd
	);

	branch_result_mux : mux
	GENERIC MAP (NBIT => NBIT)
	PORT MAP (
		s    => branch_result,
		a    => adder_plus4_result,
		b    => adder_result,
		o    => branch_mux_output
	);

	jump_imm_reg : mux
	GENERIC MAP (NBIT => NBIT)
	PORT MAP (
		s => c_j_imm_reg,
		a => jump_imm_addr,
		b => reg_rd1,
		o => jump_address
	);

	jump_result_mux : mux
	GENERIC MAP (NBIT => NBIT)
		PORT MAP (
		s => c_jump,
		a => branch_mux_output,
		b => jump_address,
		o => jump_mux_output
	);

	datapath_reg: regfile
	PORT MAP (
		clk     => clk,
		rst_s   => reset,
		we      => c_rwe,
		raddr_1 => c_raddr1,
		raddr_2 => c_raddr2,
		waddr   => c_waddr,
		rdata_1 => reg_rd1,
		rdata_2 => reg_rd2,
		wdata   => wd_result
	);

	datapath_alu : alu
	PORT MAP (
		Func_in    => c_alu,
		A_in       => alu_opA,
		B_in       => alu_opB,
		O_out      => alu_res,
		Branch_out => alu_branch
	);

	datapath_ram : dmem
	PORT MAP (
		clk   => clk,
		we    => c_mwe,
		a  => alu_res,
		wd => reg_rd2,
		rd => mem_d
	);
	
END Structural;
